#include <stdio.h>
#include <stdlib.h>

#include "configuration.h"
#include "piece.h"
#include "inputspec.h"
#include "queue.h"


extern InputSpec 	Spec;
extern int			MaxDistance;

/* Make a new, empty configuration. */

Configuration *NewConfiguration()
{
Configuration *result = (Configuration *) malloc(sizeof(Configuration));

	result->pieces = NULL;
	result->usesGoal = 0;
	
	result->distance = 0;
	result->predecessor = NULL;
	result->movedPiece = -1;
	result->movedShifts = NULL;
	
	return result;
}


/* Make a configuration which is a copy of another one. */

Configuration *CopyConfiguration(Configuration *config)
{
Configuration *result = (Configuration *) malloc(sizeof(Configuration));

	*result = *config;
	
	/* Copy the pieces. */
	
	if (result->pieces)
		result->pieces = CopyPieceList(result->pieces);
		
	/* Copy the direction list. */
	
	if (result->movedShifts)
		result->movedShifts = CopyDirectionList(result->movedShifts);
	
	return result;
}


/* Dispose the configuration contents. */

void DisposeConfiguration(Configuration *config)
{
	if (config->pieces)
	{
		DeletePLContents(config->pieces);
		free(config->pieces);
	}
	
	if (config->movedShifts)
	{
		DeleteDLContents(config->movedShifts);
		free(config->movedShifts);
	}
}


/* Try all possible piece placements at (x, y) (including 'filling' it with
   empty space), and for each one, recursively try to fill the next spot in
   the box.  If we've thus filled the entire box, we have a candidate puzzle
   layout; see if it can be solved. */

void PuzzleSearch(Configuration *config, int x, int y)
{
int nextx, nexty, i;

	if (y < Spec.boxHeight)
	{
		/* Try all piece placements at (x, y).  For i == numShapes, we choose to
		   make this mean fill square (x, y) with nothing. */
		
		nextx = (x + 1) % Spec.boxWidth;
		nexty = y + !nextx;
		
		for (i = 0; i <= Spec.numShapes; ++i)
			if (i == Spec.numShapes || PlacePiece(config, i, x, y))
			{
				PuzzleSearch(config, nextx, nexty);
				
				/* Now take the last piece out, so we can try the next. */
				
				if (i < Spec.numShapes)
					RemovePiece(config);
			}
	}
	else
		Solve(config);				/* We've filled the whole box.  Try to solve this puzzle. */
}


/* Can this configuration be solved by making some sequence of moves?
   If so, and the minimum-length solution is longer than the longest found so
   far, print it out. */

void Solve(Configuration *config)
{
Queue queue;
ConfigurationList *seen = NULL;
Configuration *config2, *newconfig;
PieceList *pl;
int pieceNum;
ConfigurationList *oldmoves, *moves;

	if (!config->usesGoal)				/* Can't solve it if it doesn't even contain the goal piece! */
		return;
	
	InitQueue(&queue);
	
	/* Do a BFS from (a copy of) this configuration, looking for a solved state.
	   See CLRS, p. 532.  We don't use an explicit 'color' field; instead, we keep
	   a list 'seen' of all configurations we've already visited. */
	
	Enqueue(&queue, CopyConfiguration(config));
		
	while (!IsEmpty(&queue))
	{
		config2 = Dequeue(&queue);

		seen = NewConfigurationList(config2, seen);			/* Add this config to the seen list. */
		
		/* Find all configurations reachable from config by moving a single piece. */
		
		for (pl = config2->pieces, pieceNum = 0; pl; pl = pl->next, ++pieceNum)
			for (moves = FindMoves(config2, pieceNum); moves; oldmoves = moves, moves = moves->next, free(oldmoves))
			{
				newconfig = moves->config;
				
				/* Have we already looked at this configuration? */
				
				if ((!seen || !IsInCL(seen, newconfig)) && !IsInQueue(&queue, newconfig))
				{
					/* Found a new reachable configuration.  Check to see if it's solved, and add it to the queue. */
					
					newconfig->distance = config2->distance + 1;
					newconfig->predecessor = config2;
					newconfig->movedPiece = pieceNum;
					
					if (Solved(newconfig))
					{							
						if (newconfig->distance > MaxDistance)
						{
							/* This is the most difficult puzzle we've found yet. */
							
							MaxDistance = newconfig->distance;

							PrintSolution(newconfig);
						}
						
						/* We know the puzzle can be solved at this depth; no use looking any farther. */
						
						DisposeConfiguration(newconfig);
						free(newconfig);
						
						goto done;
					}
					
					/* The new configuration is not solved; add it to the search queue. */
					
					Enqueue(&queue, newconfig);
				}
				else
				{	
					DisposeConfiguration(newconfig);
					free(newconfig);
				}
			}
	}
	
done:

	/* clean up */
	
	if (seen)
	{
		DeleteCLContents(seen);
		free(seen);
	}
	
	while (!IsEmpty(&queue))
	{
		config2 = Dequeue(&queue);
		DisposeConfiguration(config2);
		free(config2);
	}
}


/* Return a list of all configurations reachable from this one by a sequence of shifts of a given piece.
   To do this, we do another BFS. */

ConfigurationList *FindMoves(Configuration *config, int pieceNum)
{
Queue queue;
ConfigurationList *seen = NULL;
Configuration *start = CopyConfiguration(config);
Configuration *config2, *newconfig;
Piece *piece;
int dir;

	InitQueue(&queue);
	
	if (start->movedShifts)
	{
		DeleteDLContents(start->movedShifts);
		free(start->movedShifts);
		start->movedShifts = NULL;
	}
	
	Enqueue(&queue, start);

	while (!IsEmpty(&queue))
	{
		config2 = Dequeue(&queue);
		seen = NewConfigurationList(config2, seen);
	
		piece = PLIndex(config2->pieces, pieceNum);
		
		for (dir = 0; dir < 4; ++dir)
			if (CanSlide(piece, dir, config2))
			{
				/* We found a piece that can slide. */
				
				newconfig = CopyConfiguration(config2);
				
				Shift(PLIndex(newconfig->pieces, pieceNum), dir);
				
				/* Have we not seen this one yet? */
				
				if ((!seen || !IsInCL(seen, newconfig)) && !IsInQueue(&queue, newconfig))
				{
					/* Found a new reachable configuration.
					   Add it to the queue. */

					newconfig->movedShifts = NewDirectionList(dir, newconfig->movedShifts);
					Enqueue(&queue, newconfig);	
				}
				else
				{
					DisposeConfiguration(newconfig);
					free(newconfig);
				}
			}	
	}
	
	return seen;
}


/* Is this configuration solved?  It's solved if the goal piece is in the goal
	location. */

int	Solved(Configuration *config)
{
PieceList *pl;

	for (pl = config->pieces; pl; pl = pl->next)
		if (PieceSolved(pl->piece))
			return 1;
			
	return 0;
}


/* Print the solution.  We recursively call printSolution2 on all predecessor positions,
   until there are no more.  Then we know we're at the starting configuration,
   and can print it.  On the way back up the recursion stack, print out the
   move sequence, which we stored in the configurations. */

void PrintSolution(Configuration *config)
{
	PrintSolution2(config);

	printf("\nEND SOLUTION\n");
}

void PrintSolution2(Configuration *config)
{
	if (!config->predecessor)
	{
		/* This is our starting configuration.  Print some useful debugging & visualization stuff,
		   then print the formal solution. */
		
		printf("\n\nFound a solution!  Length is %d.  Here's what it looks like:\n", MaxDistance);

		PrettyPrint(config);

		printf("\n\nBEGIN SOLUTION\n");

		PrintPieceList(config->pieces);
		
		printf("-\n%d\n-\n", MaxDistance);
	}
	else
	{
		PrintSolution2(config->predecessor);
		
		printf("Move piece %d ", config->movedPiece);
		PrintDirectionList(config->movedShifts);
		printf("\n");
	}
}


/* Print a configuration for human consumption, in a nice ASCII grid layout. */

void PrettyPrint(Configuration *config)
{
char table[2 * MAXBOXDIM + 3][MAXBOXDIM + 2];
int x, y;
PieceList *pl;
int pieceNum;

	for (x = 0; x < 2 * MAXBOXDIM + 3; ++x)
		for (y = 0; y < MAXBOXDIM + 2; ++y)
			table[x][y] = '*';
			
	for (x = 0; x < Spec.boxWidth; ++x)
		for (y = 0; y < Spec.boxHeight; ++y)
		{
			if (!Spec.boxShape[x][y])
			{
				table[2 * x + 1][y + 1] = table[2 * x + 3][y + 1] = ' ';
				table[2 * x + 2][y + 1] = '.';
			}
			
			for (pl = config->pieces, pieceNum = 0; pl; pl = pl->next, ++pieceNum)
				if (PieceOccupies(pl->piece, x, y))
				{
					table[2 * x + 2][y + 1] = '0' + pieceNum;
					break;
				}
		}
			
	for (y = 0; y < Spec.boxHeight + 2; ++y)
	{
		for (x = 0; x < 2 * Spec.boxWidth + 3; ++x)
			printf("%c", table[x][y]);
		
		printf("\n");
	}
}


/* Can a piece of a given shape be placed into this configuration with upper
   left corner at (x, y)?
   If so, place it there & return true.  Otherwise return false. */

int PlacePiece(Configuration *config, int shapeIndex, int x, int y)
{
PieceShape *shape;
int i, j;

	/* we can only use the goal piece once. */
	
	if (shapeIndex == Spec.goalIndex && config->usesGoal)
		return 0;

	/* ... and don't bother starting it in the goal position! */
	
	if (shapeIndex == Spec.goalIndex && x == Spec.goalX && y == Spec.goalY)
		return 0;
		
	shape = &(Spec.shapes[shapeIndex]);

	for (i = 0; i < MAXPIECEDIM; ++i)
		for (j = 0; j < MAXPIECEDIM; ++j)
			if ((*shape)[i][j] && ConfigOccupies(config, x + i, y + j))
				return 0;

	config->pieces = NewPieceList(NewPiece(shapeIndex, x, y), config->pieces);
	
	if (shapeIndex == Spec.goalIndex)
		config->usesGoal = 1;
		
	return 1;
}


/* Delete the last piece added to this configuration. */

void RemovePiece(Configuration *config)
{
PieceList *oldpieces;

	if (IsGoalPiece(config->pieces->piece))
		config->usesGoal = 0;
		
	oldpieces = config->pieces;
	config->pieces = config->pieces->next;
	
	free(oldpieces);
}


/* Is the square (x, y) occupied by any piece in this configuration, or by a
 	box wall? */

int ConfigOccupies(Configuration *config, int x, int y)
{
PieceList *pl;

	if (x < 0 || y < 0 || x >= MAXBOXDIM || y >= MAXBOXDIM || Spec.boxShape[x][y])
		return 1;
		
	for (pl = config->pieces; pl; pl = pl->next)
		if (PieceOccupies(pl->piece, x, y))
			return 1;
	
	return 0;
}


/* Does this configuration have the same piece shapes and locations as another
	configuration? */

int	SameConfigAs(Configuration *config1, Configuration *config2)
{
PieceList *pl1, *pl2;

	for (pl1 = config1->pieces, pl2 = config2->pieces; pl1 || pl2; pl1 = pl1->next, pl2 = pl2->next)
	{
		if (!pl1 || !pl2)
			return 0;						/* one list was shorter than the other */
			
		if (!SamePieceAs(pl1->piece, pl2->piece))
			return 0;						/* corresponding pieces don't match */
	}
	
	return 1;
}


/* Make a new ConfigurationList node. */

ConfigurationList *NewConfigurationList(Configuration *c, ConfigurationList *cl)
{
ConfigurationList *result = (ConfigurationList *) malloc(sizeof(ConfigurationList));

	result->config = c;
	result->next = cl;
	
	return result;
}


/* Dispose the contents of the list. */

void DeleteCLContents(ConfigurationList *cl)
{
	free(cl->config);
	
	if (cl->next)
	{
		DeleteCLContents(cl->next);
		free(cl->next);
	}
}


/* Does any configuration in this list match a given configuration? */

int IsInCL(ConfigurationList *cl, Configuration *match)
{
	for (; cl; cl = cl->next)
		if (SameConfigAs(cl->config, match))
			return 1;
			
	return 0;
}


/* Make a new direction list node.
   Direction lists are used for storing the shift sequences that individual pieces do
   to make up a move. */

DirectionList *NewDirectionList(int d, DirectionList *dl)
{
DirectionList *result = (DirectionList *) malloc(sizeof(DirectionList));

	result->dir = d;
	result->next = dl;
	
	return result;
}


/* Make a copy of an existing direction list. */

DirectionList *CopyDirectionList(DirectionList *dl)
{
DirectionList *result = (DirectionList *) malloc(sizeof(DirectionList));

	result->dir = dl->dir;
	result->next = dl->next ? CopyDirectionList(dl->next) : NULL;
	
	return result;
}


/* Dispose the contents of the list. */

void DeleteDLContents(DirectionList *dl)
{
	if (dl->next)
	{
		DeleteDLContents(dl->next);
		free(dl->next);
	}
}


/* Print out the sequence of directions.  We do it in reverse order, since the
   most recent shift is at the head of the list. */

static char *DirNames[] = { "right", "down", "left", "up" };

void PrintDirectionList(DirectionList *dl)
{
	if (dl->next)
		PrintDirectionList(dl->next);
	
	printf("%s ", DirNames[dl->dir]);
}
